/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.clickhouse.reader.service;

import io.vertx.core.Future;
import io.vertx.ext.sql.SQLClient;
import io.vertx.ext.sql.SQLConnection;
import io.vertx.ext.sql.SQLRowStream;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import ru.datamart.kafka.clickhouse.reader.converter.ClickhouseFromSqlTypeConverter;
import ru.datamart.kafka.clickhouse.reader.model.DatabaseExecutorRequest;
import ru.datamart.kafka.clickhouse.reader.model.QueryResultItem;

import java.util.ArrayList;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;

@Slf4j
public class ClickhouseExecutor {

    private final SQLClient sqlClient;
    private final ClickhouseFromSqlTypeConverter typeConverter;

    public ClickhouseExecutor(SQLClient sqlClient,
                              ClickhouseFromSqlTypeConverter typeConverter) {
        this.sqlClient = sqlClient;
        this.typeConverter = typeConverter;
    }

    public Future<Void> execute(DatabaseExecutorRequest request, Queue<SQLRowStream> streams, Consumer<QueryResultItem> chunkHandler) {
        return Future.future(promise -> {
            sqlClient.getConnection(ar -> {
                if (ar.failed()) {
                    log.error("Exception during query prepare, stream number: [{}]", request.getStreamNumber(), ar.cause());
                    promise.fail(ar.cause());
                    return;
                }

                val connection = ar.result();
                receiveChunks(connection, request, streams, chunkHandler)
                        .onComplete(event -> connection.close())
                        .onSuccess(promise::complete)
                        .onFailure(promise::fail);
            });
        });
    }

    private Future<Void> receiveChunks(SQLConnection connection,
                                       DatabaseExecutorRequest request,
                                       Queue<SQLRowStream> streams,
                                       Consumer<QueryResultItem> chunkHandler) {
        return Future.future(promise -> {
            log.info("Start query execution [{}], stream number: [{}]", request.getSql(), request.getStreamNumber());
            connection.queryStream(request.getSql(), ar -> {
                if (ar.failed()) {
                    log.error("Exception during query prepare, stream number: [{}]", request.getStreamNumber(), ar.cause());
                    promise.fail(ar.cause());
                    return;
                }

                val chunkNumber = new AtomicInteger(1);
                val dataSet = new ArrayList<List<?>>();
                val stream = ar.result();
                streams.add(stream);
                stream.exceptionHandler(t -> {
                    log.error("Exception during stream read, stream number: [{}]", request.getStreamNumber(), t);
                    closeStream(dataSet, stream);
                    promise.fail(t);
                }).endHandler(v -> {
                    try {
                        log.info("Stream read success, last chunk: [{}], stream number: [{}]", chunkNumber.get(), request.getStreamNumber());
                        chunkHandler.accept(new QueryResultItem(new ArrayList<>(dataSet), chunkNumber.getAndIncrement(), true));
                        closeStream(dataSet, stream);
                        promise.complete();
                    } catch (Exception e) {
                        log.error("Exception during stream end handler, stream number: [{}]", request.getStreamNumber(), e);
                        closeStream(dataSet, stream);
                        promise.fail(e);
                    }
                }).handler(row -> {
                    try {
                        val rowData = new ArrayList<>(row.size());
                        for (int i = 0; i < row.size(); i++) {
                            rowData.add(typeConverter.convert(request.getColumnTypes().get(i), row.getValue(i)));
                        }
                        dataSet.add(rowData);
                        if (dataSet.size() == request.getChunkSize()) {
                            log.debug("Stream read chunk [{}] filled, stream number: [{}]", chunkNumber.get(), request.getStreamNumber());
                            chunkHandler.accept(new QueryResultItem(new ArrayList<>(dataSet), chunkNumber.getAndIncrement(), false));
                            dataSet.clear();
                        }
                    } catch (Exception e) {
                        log.error("Exception during stream handle, stream number: [{}]", request.getStreamNumber(), e);
                        closeStream(dataSet, stream);
                        promise.fail(e);
                    }
                });
            });
        });
    }

    private void closeStream(ArrayList<List<?>> dataSet, SQLRowStream stream) {
        dataSet.clear();
        stream.close();
    }
}
