/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.clickhouse.reader.verticle;

import io.vertx.core.DeploymentOptions;
import io.vertx.core.Future;
import io.vertx.ext.sql.SQLRowStream;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import org.apache.avro.Schema;
import ru.datamart.kafka.clickhouse.reader.configuration.properties.ReaderProperties;
import ru.datamart.kafka.clickhouse.reader.configuration.properties.VerticleProperties.WorkerProperties;
import ru.datamart.kafka.clickhouse.reader.model.Chunk;
import ru.datamart.kafka.clickhouse.reader.model.DatabaseExecutorRequest;
import ru.datamart.kafka.clickhouse.reader.model.ReaderProcessContext;
import ru.datamart.kafka.clickhouse.reader.service.ClickhouseExecutor;

import java.util.ArrayList;
import java.util.List;


@Slf4j
public class ClickhouseExecutorVerticle extends ConfigurableVerticle {
    private final ReaderProcessContext context;
    private final ClickhouseExecutor clickhouseExecutor;
    private final ReaderProperties readerProperties;
    private final WorkerProperties properties;
    private volatile long timerId = -1;
    private final int streamNumber;
    private final int streamTotal;

    public ClickhouseExecutorVerticle(ReaderProcessContext context,
                                      ClickhouseExecutor clickhouseExecutor,
                                      ReaderProperties readerProperties,
                                      WorkerProperties properties,
                                      int streamNumber,
                                      int streamTotal) {
        this.context = context;
        this.clickhouseExecutor = clickhouseExecutor;
        this.readerProperties = readerProperties;
        this.properties = properties;
        this.streamNumber = streamNumber;
        this.streamTotal = streamTotal;
    }

    @Override
    public DeploymentOptions getDeploymentOptions() {
        return new DeploymentOptions()
                .setWorker(true)
                .setWorkerPoolName("clickhouse-worker-pool")
                .setWorkerPoolSize(properties.getPoolSize());
    }

    @Override
    public void start() {
        vertx.eventBus().consumer(context.getStartAddress(), ignored -> execute()
                .onSuccess(event -> {
                    log.info("Query execute SUCCESS");
                    context.updateStateFinishOneExecution();
                })
                .onFailure(throwable -> {
                    log.error("Query execute FAILED", throwable);
                    context.updateStateSetError(throwable);
                }));
    }

    @Override
    public void stop() {
        vertx.cancelTimer(timerId);
    }

    private Future<Void> execute() {
        timerId = vertx.setPeriodic(500, event -> {
            if (context.getUploadingCount() >= readerProperties.getUploadChunkBuffer() && context.getPausedStreams().compareAndSet(false, true)) {
                context.getStreams().forEach(SQLRowStream::pause);
                log.debug("Paused query read stream, chunk buffer filled [{}]", readerProperties.getUploadChunkBuffer());
            }

            if (context.getUploadingCount() <= readerProperties.getUploadChunkBuffer() / 10 && context.getPausedStreams().compareAndSet(true, false)) {
                context.getStreams().forEach(SQLRowStream::resume);
                log.debug("Resumed query read stream, chunk buffer freed [{}]", readerProperties.getUploadChunkBuffer());
            }
        });

        return Future.future(promise -> {
            val queryRequest = context.getQueryRequest();
            val columnTypes = getColumnTypes();
            val request = DatabaseExecutorRequest.builder()
                    .streamNumber(streamNumber)
                    .streamTotal(streamTotal)
                    .chunkSize(queryRequest.getChunkSize())
                    .sql(queryRequest.getSql())
                    .columnTypes(columnTypes)
                    .build();

            clickhouseExecutor.execute(request, context.getStreams(), chunk -> {
                                val error = context.getState().getError();
                                if (error != null) {
                                    throw new IllegalStateException("Stopping database stream because of process exception", error);
                                }

                                context.incrementUploadingCount();
                                context.getChunksToUpload().add(new Chunk(chunk, streamNumber, streamTotal));
                                vertx.eventBus().send(context.getProcessChunkAddress(), null);
                            }
                    )
                    .onComplete(promise);
        });
    }

    private List<Schema.Type> getColumnTypes() {
        val schema = new Schema.Parser().parse(context.getQueryRequest().getAvroSchema());
        val fields = schema.getFields();
        List<Schema.Type> result = new ArrayList<>();
        for (val field : fields) {
            val typeSchema = field.schema();
            if (typeSchema.isUnion()) {
                val first = typeSchema.getTypes().stream()
                        .filter(schema2 -> schema2.getType() != Schema.Type.NULL)
                        .findFirst();
                if (first.isPresent()) {
                    result.add(first.get().getType());
                } else {
                    result.add(Schema.Type.NULL);
                }
                continue;
            }

            result.add(typeSchema.getType());
        }
        return result;
    }

}
