/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.clickhouse.reader.verticle;

import io.vertx.core.Future;
import io.vertx.core.Promise;
import io.vertx.core.Vertx;
import io.vertx.ext.sql.SQLRowStream;
import io.vertx.junit5.VertxExtension;
import io.vertx.junit5.VertxTestContext;
import lombok.val;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import ru.datamart.kafka.clickhouse.reader.configuration.properties.ReaderProperties;
import ru.datamart.kafka.clickhouse.reader.configuration.properties.VerticleProperties.WorkerProperties;
import ru.datamart.kafka.clickhouse.reader.model.QueryRequest;
import ru.datamart.kafka.clickhouse.reader.model.ReaderProcessContext;
import ru.datamart.kafka.clickhouse.reader.service.ClickhouseExecutor;

import java.util.Arrays;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicBoolean;

import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

@ExtendWith({VertxExtension.class, MockitoExtension.class})
class ClickhouseExecutorVerticleTest {
    private static final String SCHEMA = "{\n" +
            "    \"type\": \"record\",\n" +
            "    \"name\": \"dwnl_test\",\n" +
            "    \"namespace\": \"pva_pxf_test\",\n" +
            "    \"fields\": [\n" +
            "      {\n" +
            "        \"name\": \"id\",\n" +
            "        \"type\": [\n" +
            "          \"null\",\n" +
            "          \"long\"\n" +
            "        ],\n" +
            "        \"default\": null\n" +
            "      }\n" +
            "    ]\n" +
            "  }";

    public static final String START_ADDRESS = "start";
    @Mock
    private ReaderProcessContext context;
    @Mock
    private ClickhouseExecutor clickhouseExecutor;
    @Mock
    private ReaderProperties readerProperties;
    @Mock
    private WorkerProperties properties;
    @Mock
    private SQLRowStream stream;

    private volatile String verticle;

    @BeforeEach
    void setUp(Vertx vertx, VertxTestContext testContext) {
        lenient().when(readerProperties.getUploadChunkBuffer()).thenReturn(1000L);
        lenient().when(context.getUploadingCount()).thenReturn(0L);
        lenient().when(context.getStartAddress()).thenReturn(START_ADDRESS);
        QueryRequest value = new QueryRequest();
        value.setAvroSchema(SCHEMA);
        lenient().when(context.getQueryRequest()).thenReturn(value);
        val streams = new ConcurrentLinkedQueue<>(Arrays.asList(stream));
        lenient().when(context.getStreams()).thenReturn(streams);
        val pausedStreams = new AtomicBoolean(false);
        lenient().when(context.getPausedStreams()).thenReturn(pausedStreams);

        vertx.deployVerticle(new ClickhouseExecutorVerticle(context, clickhouseExecutor, readerProperties, properties, 0, 1))
                .onSuccess(deploymentId -> {
                    verticle = deploymentId;
                    testContext.completeNow();
                })
                .onFailure(testContext::failNow);
    }

    @AfterEach
    void cleanUp(Vertx vertx) {
        vertx.undeploy(verticle);
    }

    @Test
    void shouldSuccessWhenEverythingOk(VertxTestContext testContext, Vertx vertx) {
        // arrange
        when(clickhouseExecutor.execute(any(), any(), any())).thenReturn(Future.succeededFuture());

        // act
        vertx.eventBus().send(START_ADDRESS, null);

        // assert
        testContext.verify(() -> {
            verify(clickhouseExecutor, timeout(1000L).times(1)).execute(any(), any(), any());
            verify(context, timeout(1000L).times(1)).updateStateFinishOneExecution();
        }).completeNow();
    }

    @Test
    void shouldStopAndResumeStreamWhenChunkBufferFilled(VertxTestContext testContext, Vertx vertx) {
        // arrange
        when(context.getUploadingCount()).thenReturn(1000L);

        Promise<Void> promise = Promise.promise();
        when(clickhouseExecutor.execute(any(), any(), any())).thenReturn(promise.future());

        // act
        vertx.eventBus().send(START_ADDRESS, null);

        // assert
        testContext.verify(() -> {
            verify(stream, timeout(1000L).times(1)).pause();
            assertTrue(context.getPausedStreams().get());

            when(context.getUploadingCount()).thenReturn(0L);
            verify(stream, timeout(1000L).times(1)).resume();
            assertFalse(context.getPausedStreams().get());
        }).completeNow();
    }

    @Test
    void shouldFailWhenExecutionFailed(VertxTestContext testContext, Vertx vertx) {
        // arrange
        val exception = new RuntimeException("Exception");
        when(clickhouseExecutor.execute(any(), any(), any())).thenReturn(Future.failedFuture(exception));

        // act
        vertx.eventBus().send(START_ADDRESS, null);

        // assert
        testContext.verify(() -> {
            verify(clickhouseExecutor, timeout(1000L).times(1)).execute(any(), any(), any());

            verify(context, timeout(1000L).times(1)).updateStateSetError(same(exception));
        }).completeNow();
    }
}