/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.clickhouse.writer.model.kafka;

import io.vertx.core.Future;
import io.vertx.kafka.client.common.TopicPartition;
import io.vertx.kafka.client.consumer.KafkaConsumer;
import io.vertx.kafka.client.consumer.OffsetAndMetadata;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import lombok.var;

import java.util.HashMap;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicLong;

@Data
@Slf4j
public class TopicPartitionConsumer {
    private final Queue<Long> notCommittedOffsets = new ConcurrentLinkedQueue<>();
    private final KafkaConsumer<byte[], byte[]> kafkaConsumer;
    private final TopicPartition topicPartition;
    private AtomicLong lastOffset;

    public TopicPartitionConsumer(KafkaConsumer<byte[], byte[]> kafkaConsumer,
                                  TopicPartition topicPartition,
                                  long lastOffset) {
        this.kafkaConsumer = kafkaConsumer;
        this.topicPartition = topicPartition;
        this.lastOffset = new AtomicLong(lastOffset);
    }

    public Future<Void> commitCompletedOffsets() {
        return Future.future(promise -> {
            var committedOffset = lastOffset.get();
            log.debug("TopicPartition[{}]: lastOffset[{}]", topicPartition, committedOffset);

            long newOffset = committedOffset;
            while (!notCommittedOffsets.isEmpty()) {
                val offset = notCommittedOffsets.poll();
                if (offset == null) {
                    continue;
                }

                long offsetToBeCommitted = offset + 1;
                if (offsetToBeCommitted > newOffset) {
                    newOffset = offsetToBeCommitted;
                }
            }

            if (newOffset == committedOffset) {
                log.debug("TopicPartition[{}]: lastOffset[{}] - notCommittedOffsets[{}]", topicPartition, committedOffset, notCommittedOffsets);
                promise.complete();
                return;
            }

            lastOffset.set(newOffset);
            log.debug("TopicPartition[{}]: committedOffset[{}], lastOffset[{}]", topicPartition, committedOffset, newOffset);
            val offsetMap = new HashMap<TopicPartition, OffsetAndMetadata>();
            offsetMap.put(topicPartition, new OffsetAndMetadata(newOffset, null));
            kafkaConsumer.commit(offsetMap)
                    .<Void>mapEmpty()
                    .onSuccess(ignored -> log.debug("Message commit for topic [{}] was successful. Offsets[{}]", topicPartition.getTopic(), offsetMap))
                    .onFailure(t -> log.debug("Message commit error for topic [{}]: [{}]", topicPartition.getTopic(), t))
                    .onComplete(promise);
        });
    }

    public void addCompletedOffset(long offset) {
        notCommittedOffsets.add(offset);
    }
}
