/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.clickhouse.writer.service.kafka;

import io.vertx.core.Future;
import io.vertx.kafka.client.common.PartitionInfo;
import io.vertx.kafka.client.common.TopicPartition;
import io.vertx.kafka.client.consumer.KafkaConsumer;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import org.apache.kafka.clients.consumer.ConsumerConfig;
import org.springframework.stereotype.Service;
import ru.datamart.kafka.clickhouse.writer.configuration.properties.kafka.KafkaProperties;
import ru.datamart.kafka.clickhouse.writer.factory.impl.VertxKafkaConsumerFactory;
import ru.datamart.kafka.clickhouse.writer.model.InsertDataContext;
import ru.datamart.kafka.clickhouse.writer.model.kafka.KafkaBrokerInfo;
import ru.datamart.kafka.clickhouse.writer.model.kafka.TopicPartitionConsumer;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static java.util.stream.Collectors.toList;

@Slf4j
@Service
public class KafkaConsumerService {
    private static final String BOOTSTRAP_SERVERS_KEY = "bootstrap.servers";
    private final VertxKafkaConsumerFactory<byte[], byte[]> consumerFactory;
    private final KafkaProperties kafkaProperties;

    public KafkaConsumerService(VertxKafkaConsumerFactory<byte[], byte[]> consumerFactory,
                                KafkaProperties kafkaProperties) {
        this.consumerFactory = consumerFactory;
        this.kafkaProperties = kafkaProperties;
    }

    public Future<List<PartitionInfo>> getTopicPartitions(InsertDataContext context) {
        return Future.future(p -> {
            val consumerProperty = getConsumerProperty(context);
            val topicName = context.getRequest().getKafkaTopic();
            val partitionInfoConsumer = createConsumer(consumerProperty, topicName);
            partitionInfoConsumer.partitionsFor(topicName)
                    .onComplete(ar -> {
                        partitionInfoConsumer.close();
                        if (ar.succeeded()) {
                            p.complete(ar.result());
                        } else {
                            val errMsg = String.format("Can't get partition info by topic [%s]", topicName);
                            log.error(errMsg, ar.cause());
                            p.fail(ar.cause());
                        }
                    });
        });
    }

    private KafkaConsumer<byte[], byte[]> createConsumer(Map<String, String> consumerProperty,
                                                         String topicName) {
        val consumer = consumerFactory.create(consumerProperty);
        consumer.subscribe(topicName, it -> {
            if (it.succeeded()) {
                log.debug("Successful topic subscription {}", topicName);
            } else {
                log.error("Topic Subscription Error {}: {}", topicName, it.cause());
            }
        });
        consumer.exceptionHandler(it -> log.error("Error reading message from topic {}", topicName, it));
        return consumer;
    }

    public Future<TopicPartitionConsumer> createTopicPartitionConsumer(InsertDataContext context, PartitionInfo partitionInfo) {
        return Future.future(promise -> {
            val consumer = consumerFactory.create(getConsumerProperty(context));
            consumer.exceptionHandler(it -> log.error("Error reading message from TopicPartition {}", partitionInfo, it));
            val topicPartition = new TopicPartition(partitionInfo.getTopic(), partitionInfo.getPartition());
            consumer.assign(topicPartition)
                    .compose(v -> consumer.position(topicPartition))
                    .map(beginningOffset -> new TopicPartitionConsumer(consumer, topicPartition, beginningOffset))
                    .onSuccess(topicPartitionConsumer -> log.info("Created consumer[{}] by topicPartition [{}]", topicPartitionConsumer, topicPartition))
                    .onComplete(promise);
        });
    }

    private HashMap<String, String> getConsumerProperty(InsertDataContext context) {
        val props = new HashMap<>(kafkaProperties.getConsumer().getProperty());
        props.put(ConsumerConfig.GROUP_ID_CONFIG, context.getRequest().getConsumerGroup());
        List<String> brokerAddresses = context.getRequest().getKafkaBrokers().stream()
                .map(KafkaBrokerInfo::getAddress).collect(toList());
        props.put(BOOTSTRAP_SERVERS_KEY, String.join(",", brokerAddresses));
        return props;
    }
}
